project = "Nebari CLI documentation"
copyright = "2023, Nebari"
author = "Nebari team"

extensions = ["sphinx_click"]

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

html_theme = "alabaster"
